"""
This module evaluates the parsed condition expression tree regarding the requirement constraints using ConditionNodes.
The RequirementConstraintTransformer defines the rules how the different parts and nodes
of the condition expression tree are handled.

The used terms are defined in the README_conditions.md.
"""

from typing import List, Literal, Mapping, Optional, Type, Union

from lark import Token, Tree, v_args
from lark.exceptions import VisitError

from ahbicht.condition_node_builder import ConditionNodeBuilder, TRCTransformerArgument
from ahbicht.evaluation_results import RequirementConstraintEvaluationResult
from ahbicht.expressions import InvalidExpressionError
from ahbicht.expressions.base_transformer import BaseTransformer
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree
from ahbicht.expressions.condition_nodes import (
    ConditionFulfilledValue,
    ConditionNode,
    EvaluatedComposition,
    FormatConstraint,
    Hint,
    RequirementConstraint,
    UnevaluatedFormatConstraint,
)
from ahbicht.expressions.expression_builder import FormatConstraintExpressionBuilder, HintExpressionBuilder


@v_args(inline=True)  # Children are provided as *args instead of a list argument
# pylint:disable=inherit-non-class
class RequirementConstraintTransformer(BaseTransformer[TRCTransformerArgument, EvaluatedComposition]):
    """
    Transformer that evaluates the trees built from the condition expressions regarding the requirement constraints.
    The input are the conditions as defined in the AHBs in the form of ConditionNodes.
    In the case of a Bedingung/RequirementConstraint already evaluated for the single condition.
    Hints are evaluated as neutral element in the boolean logic: and_composition: True, or_composition: False.

    After two nodes are evaluated in a composition, the resulting node has the node type EvalutatedComposition.
    The return value is a ConditionNode whose attribute `conditions_fulfilled` describes whether it is a required field,
    its hints are the gathered hints and its `format_constraint_expression` contains the gathered format constraints
    for the field in a newly built expression.
    """

    def and_composition(self, left: TRCTransformerArgument, right: TRCTransformerArgument) -> EvaluatedComposition:
        """Evaluates logical and_composition"""

        # if one of the nodes is neutral the condition_fulfilled.value of the other one is the resulting one
        evaluated_composition = EvaluatedComposition(
            conditions_fulfilled=left.conditions_fulfilled & right.conditions_fulfilled
        )

        # Hints are added if the branch is true or neutral
        if evaluated_composition.conditions_fulfilled != ConditionFulfilledValue.UNFULFILLED:
            evaluated_composition.hint = (
                HintExpressionBuilder(getattr(left, "hint", None)).land(getattr(right, "hint", None)).get_expression()
            )
        evaluated_composition.format_constraints_expression = (
            FormatConstraintExpressionBuilder(left).land(right).get_expression()
        )

        return evaluated_composition

    def _or_xor_composition(
        self, left: ConditionNode, right: ConditionNode, composition: Literal["or_composition", "xor_composition"]
    ):
        """
        Determine the condition_fulfilled attribute for or_/xor_compostions.
        """
        # combining a hint with a format constraint in an or_/xor_composition has no useful result
        if (isinstance(left, Hint) and isinstance(right, UnevaluatedFormatConstraint)) or (
            isinstance(right, Hint) and isinstance(left, UnevaluatedFormatConstraint)
        ):
            raise InvalidExpressionError(
                error_message=f"Combining a {left.__class__} and a {right.__class__} in an"
                f"{composition} is not implemented as it has no useful result."
            )

        # combining a neutral element with a boolean value in an or_/xor_composition has no useful result
        if (
            left.conditions_fulfilled == ConditionFulfilledValue.NEUTRAL
            and right.conditions_fulfilled != ConditionFulfilledValue.NEUTRAL
            or (
                right.conditions_fulfilled == ConditionFulfilledValue.NEUTRAL
                and left.conditions_fulfilled != ConditionFulfilledValue.NEUTRAL
            )
        ):
            neutral_element: ConditionNode
            boolean_element: ConditionNode
            if right.conditions_fulfilled == ConditionFulfilledValue.NEUTRAL:
                neutral_element = right
                boolean_element = left
            else:
                neutral_element = left
                boolean_element = right
            error_message = (
                f"Combining a neutral element '{neutral_element}' with a boolean value {boolean_element} "
                f"in an {composition} is not implemented as it has no useful result."
            )
            if isinstance(neutral_element, Hint):
                # This error is quite popular in AHBs.
                # The BDEW tends to write stuff like "[123] O [584]" where [123] is a regular RC and [584] is
                # something like "Verwendung der ID der Markt/Messlokation". Their intention is that the hint on the
                # right side should be evaluated as if it was a self-referencing Requirement Constraint that
                # magically evaluates to Fulfilled/Unfulfilled instead of being treated as what it is: A plain hint.
                # This is similar to what they tried with [93x] (german datetime) format constraints.
                # But other than the 931-935 FCs this kind of expression is not only a bad idea but also not even
                # valid according to their own set of rules.
                # If you encounter such an error in the AHBs, please message the EDI@Energy Arbeitsgruppe to fix it.
                # The easiest workaround (beyond ignoring it) is probably to convert it to an _xor_ expression:
                # Instead of "[123] O [584]" write "[983][123] X [984][1]" where
                # - [1] is an artificial requirement constraint, that should always evaluate to FULFILLED
                #   (Note that the "[1]" does not exist by default; Here it's just a placeholder for this example.)
                # - [984] is the proper format constraint that has the same meaning as the hint [584]
                # - [983] is the complementary FC to [984] ("not [984]")
                error_message += " This is probably an error in the AHB itself."
            raise InvalidExpressionError(error_message)
        if composition == "or_composition":
            resulting_conditions_fulfilled = left.conditions_fulfilled | right.conditions_fulfilled
        elif composition == "xor_composition":
            resulting_conditions_fulfilled = left.conditions_fulfilled ^ right.conditions_fulfilled
        # resulting_conditions_fulfilled is never referenced before assignment when you obey the type hint
        evaluated_composition = EvaluatedComposition(conditions_fulfilled=resulting_conditions_fulfilled)
        return evaluated_composition

    def or_composition(self, left: ConditionNode, right: ConditionNode) -> EvaluatedComposition:
        """Evaluates logical (inclusive) or_composition"""

        evaluated_composition = self._or_xor_composition(left, right, "or_composition")

        evaluated_composition.format_constraints_expression = (
            # todo: ask annika why the case of "invalid arguments" never happens here...
            FormatConstraintExpressionBuilder(left)  # type:ignore[arg-type]
            .lor(right)  # type:ignore[arg-type]
            .get_expression()
        )
        evaluated_composition.hint = (
            HintExpressionBuilder(getattr(left, "hint", None)).lor(getattr(right, "hint", None)).get_expression()
        )

        return evaluated_composition

    def xor_composition(self, left: ConditionNode, right: ConditionNode) -> EvaluatedComposition:
        """Evaluates exclusive xor_composition"""

        evaluated_composition = self._or_xor_composition(left, right, "xor_composition")

        evaluated_composition.format_constraints_expression = (
            # todo: ask annika why the case of "invalid arguments" never happens here...
            FormatConstraintExpressionBuilder(left)  # type:ignore[arg-type]
            .xor(right)  # type:ignore[arg-type]
            .get_expression()
        )
        evaluated_composition.hint = (
            HintExpressionBuilder(getattr(left, "hint", None)).xor(getattr(right, "hint", None)).get_expression()
        )

        return evaluated_composition

    def _then_also(
        self,
        format_constraint: UnevaluatedFormatConstraint,
        other_condition: Type[ConditionNode],
    ) -> EvaluatedComposition:
        """
        Evaluates a boolean condition with a format constraint. The functions name indicates its behaviour:
        If the `condition_fulfilled` attribute is true, the format constraint has to be fulfilled, too.
        But it is not directly evaluated but only attached to the evaluated composition such that we keep
        a strict separation of the Mussfeldprüfung itself and evaluation of format constraints on fields
        that have been marked obligatory by the (previously run) Mussfeldprüfung but do not affect
        the result of the Mussfeldprüfung.
        (Things that are obligatory, are obligatory regardless of the format constraints.)

        :param format_constraint: a format constraint
        :param other_condition: a requirement constraint or hint
        :return:
        """
        if other_condition.conditions_fulfilled != ConditionFulfilledValue.NEUTRAL:
            evaluated_composition = EvaluatedComposition(conditions_fulfilled=other_condition.conditions_fulfilled)
            format_constraint_is_required = other_condition.conditions_fulfilled == ConditionFulfilledValue.FULFILLED
        elif isinstance(other_condition, Hint):
            evaluated_composition = EvaluatedComposition(
                conditions_fulfilled=ConditionFulfilledValue.NEUTRAL, hint=other_condition.hint
            )
            format_constraint_is_required = True
        else:
            raise NotImplementedError(
                f"Combining Conditions of type {other_condition.__class__} with a "
                f"{format_constraint.__class__} is not implemented."
            )
        if format_constraint_is_required:
            evaluated_composition.format_constraints_expression = (
                FormatConstraintExpressionBuilder(format_constraint).land(other_condition).get_expression()
            )

        return evaluated_composition

    def then_also_composition(self, left: Type[ConditionNode], right: Type[ConditionNode]) -> EvaluatedComposition:
        """
        A "then also" composition is typically used for format constraints.
        It connects an evaluable expression with a format constraint.
        If the `conditions_fulfilled` attribute evaluates to true,
        then also the attached format constraint has to be fulfilled.
        """
        # pylint: disable=protected-access
        if isinstance(left, UnevaluatedFormatConstraint):
            return self._then_also(format_constraint=left, other_condition=right)
        return self._then_also(
            format_constraint=right,  # type:ignore[arg-type]
            # #because typically format constraints are "attached" to the right side of an expression
            other_condition=left,
        )  # this might raise the NotImplementedError


def evaluate_requirement_constraint_tree(
    parsed_tree: Tree, input_values: Mapping[str, TRCTransformerArgument]
) -> EvaluatedComposition:
    """
    Evaluates the tree built from the expressions with the help of the ConditionsTransformer.

    :param parsed_tree: Tree
    :param input_values: dict(condition_key, ConditionNode)
        :return: EvaluatedComposition
    """
    if not all(isinstance(input_value, ConditionNode) for input_value in input_values.values()):
        raise ValueError("Please make sure that the passed values are ConditionNodes.")
    if not all(
        isinstance(input_value, (RequirementConstraint, Hint, FormatConstraint))
        for input_value in input_values.values()
    ):
        raise ValueError(
            """Please make sure that the passed values are ConditionNodes \
of the type RequirementConstraint, Hint or FormatConstraint."""
        )
    try:
        result = RequirementConstraintTransformer(input_values).transform(parsed_tree)
    except VisitError as visit_err:
        raise visit_err.orig_exc

    return result


async def requirement_constraint_evaluation(
    condition_expression: Union[str, Tree]
) -> RequirementConstraintEvaluationResult:
    """
    Evaluation of the condition expression in regard to the requirement conditions (rc).
    The condition expression can either be a string that still needs to be parsed as condition expression or a tree
    that has already been parsed.
    """
    if isinstance(condition_expression, str):
        parsed_tree_rc: Tree = parse_condition_expression_to_tree(condition_expression)
    else:
        parsed_tree_rc = condition_expression

    # get all condition keys from tree
    all_condition_keys: List[str] = [
        t.value for t in parsed_tree_rc.scan_values(lambda v: isinstance(v, Token))  # type: ignore[attr-defined]
    ]
    condition_node_builder = ConditionNodeBuilder(all_condition_keys)
    input_nodes = await condition_node_builder.requirement_content_evaluation_for_all_condition_keys()

    resulting_condition_node: EvaluatedComposition = evaluate_requirement_constraint_tree(parsed_tree_rc, input_nodes)

    requirement_constraints_fulfilled: Optional[bool] = (
        resulting_condition_node.conditions_fulfilled == ConditionFulfilledValue.FULFILLED
    )
    requirement_is_conditional: Optional[bool] = True
    if resulting_condition_node.conditions_fulfilled == ConditionFulfilledValue.NEUTRAL:  # pylint:disable=no-member
        requirement_constraints_fulfilled = True
        requirement_is_conditional = False
    if resulting_condition_node.conditions_fulfilled == ConditionFulfilledValue.UNKNOWN:  # pylint:disable=no-member
        # unknown_keys = [
        #    node_key
        #    for node_key, node_value in input_nodes.items()
        #    if isinstance(node_value, RequirementConstraint)
        #    and node_value.conditions_fulfilled == ConditionFulfilledValue.UNKNOWN
        # ]
        # a NotImplementedError was raised here in ahbicht<=v0.5.11
        # https://github.com/Hochfrequenz/ahbicht/issues/275
        requirement_constraints_fulfilled = None
        requirement_is_conditional = None

    format_constraints_expression = getattr(resulting_condition_node, "format_constraints_expression", None)
    if isinstance(resulting_condition_node, UnevaluatedFormatConstraint):
        format_constraints_expression = f"[{resulting_condition_node.condition_key}]"
    hints = getattr(resulting_condition_node, "hint", None)

    requirement_constraint_evaluation_result = RequirementConstraintEvaluationResult(
        requirement_constraints_fulfilled=requirement_constraints_fulfilled,
        hints=hints,
        format_constraints_expression=format_constraints_expression,
        requirement_is_conditional=requirement_is_conditional,
    )

    return requirement_constraint_evaluation_result



"""
Tests the hints provider module.
"""
import asyncio
import datetime
from logging import LogRecord
from typing import List

import pytest  # type:ignore[import]
from maus.edifact import EdifactFormat, EdifactFormatVersion

from ahbicht.expressions.hints_provider import HintsProvider, JsonFileHintsProvider


class Dummy1sHintsProvider(HintsProvider):
    """a hints provider that takes 1s for each (dummy) hint text"""

    async def get_hint_text(self, _: str):
        await asyncio.sleep(1)
        return "foo"


class DummyAsyncHintsProvider(HintsProvider):
    """a hints provider that has an async get_hint_text_method"""

    async def get_hint_text(self, _: str):
        return "foo"


class DummySyncHintsProvider(HintsProvider):
    """a hints provider that has a sync get_hint_text_method"""

    def get_hint_text(self, _: str):
        return "foo"


class TestHintsProvider:
    """Test Class for JsonFileHintsProvider"""

    @pytest.mark.datafiles("./unittests/provider_test_files/example_hints_file.json")
    async def test_initiating_hints_provider(self, datafiles):
        """Tests if hints provider is initiated correctly."""
        path_to_hint_json = datafiles / "example_hints_file.json"
        hints_provider = JsonFileHintsProvider(
            edifact_format=EdifactFormat.UTILMD,
            edifact_format_version=EdifactFormatVersion.FV2104,
            file_path=path_to_hint_json,
        )
        assert hints_provider.edifact_format == EdifactFormat.UTILMD
        assert hints_provider.edifact_format_version == EdifactFormatVersion.FV2104
        assert await hints_provider.get_hint_text("583") == "[583] Hinweis: Verwendung der ID der Marktlokation"

    async def test_concurrent_hint_text_resolving(self):
        """
        Tests that the get_hint_text methods are evaluated concurrently
        :return:
        """
        hints_provider = Dummy1sHintsProvider()
        dummy_keys = ["1", "2", "3"]
        start = datetime.datetime.now()
        await hints_provider.get_hints(dummy_keys)
        end = datetime.datetime.now()
        assert (end - start).total_seconds() < len(dummy_keys)

    async def test_sync_hint_text_resolving(self, caplog):
        hints_provider = DummySyncHintsProvider()
        dummy_keys = ["1", "2", "3"]
        await hints_provider.get_hints(dummy_keys)
        log_entries: List[LogRecord] = caplog.records
        assert len(log_entries) == 2
        assert log_entries[0].message == "Instantiated DummySyncHintsProvider"
        assert log_entries[1].message == "Found 3 hints for 1, 2, 3"

    async def test_async_hint_text_resolving(self):
        hints_provider = DummyAsyncHintsProvider()
        dummy_keys = ["1", "2", "3"]
        await hints_provider.get_hints(dummy_keys)



"""
A format constraint (FC) evaluator is an evaluator for format constraints.
Think of stuff like:
* strings that should match an OBIS regex
* MarktlokationsIDs having correct check sums
* pre/post decimal values having specific ranges
Other than requirement constraints format constraints do not affect if data are required at all, but instead only
validate already required data.
"""
import asyncio
import inspect
from abc import ABC
from contextvars import ContextVar
from typing import Callable, Coroutine, Dict, List, Optional

import inject

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult, ContentEvaluationResultSchema
from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData, EvaluatableDataProvider
from ahbicht.content_evaluation.evaluators import Evaluator
from ahbicht.content_evaluation.german_strom_and_gas_tag import has_no_utc_offset, is_xtag_limit
from ahbicht.evaluation_results import FormatConstraintEvaluationResult
from ahbicht.expressions.condition_nodes import EvaluatedFormatConstraint

text_to_be_evaluated_by_format_constraint: ContextVar[Optional[str]] = ContextVar(
    "text_to_be_evaluated_by_format_constraint", default=None
)
"""
This context variable holds the text that is to be analysed/evaluated by the format constraint evaluator.
It will always return the "correct" value in your context. You only have to manually set this context variable if you're
evaluating an expression outside the validation framework.
The conceptual difference to the EvaluatableData which are dependency injected using the EvaluatableDataProvider is,
that the data evaluated in a format constraint (via the context variable) vary over the time span of one validation run.
The EvaluatableData are stable in that regard.
"""

# The idea behind the context variable is to avoid passing the string/text to be evaluated by FC evaluators through many
# layers of code (including lark classes like transformers where we formerly had the string as constructor argument).
# Now it needs to be set once and only once when the entered input is determined (e.g. when reading user input).
# Then we can forget about it, it does not bloat our function signatures (where in >90% of the cases it has just been
# forwarded to the next layer of code).
# The context variable also greatly simplifies the debugging/error analysis.
# Instead of tracing the entered input all the way down from the ahbicht API to the FC Evaluator method,
# you now just need to watch the two places in the code where the values are actually set:
# 1. in the ahbicht code base in the validation
# 2. in custom code using ahbicht(?)
# The single evaluation methods of the FC evaluators (e.g. "def evaluate_987") are still provided with the value as a
# usual function argument to prevent the methods from needing to access the context variable themselves.


class FcEvaluator(Evaluator, ABC):
    """
    Base of all Format Constraint (FC) evaluators.
    To implement a content_evaluation create or update an inheriting class that has edifact_format and
    edifact_format_version set accordingly. Then create a method named "evaluate_123" where "123" is the condition key
    of the condition it evaluates.
    """

    def evaluate_931(self, entered_input: str) -> EvaluatedFormatConstraint:
        """
        Assert that the entered input is parsable as datetime with explicit UTC offset.
        Then assert that the UTC offset is exactly +00:00.
        Be aware of the fact that asserting on a fixed offset when both datetime + offset are given will not lead to any
        truly meaningful results.
        We implement it for compatability but don't encourage you to actively write any conditions that use it.
        """
        return has_no_utc_offset(entered_input)

    def evaluate_932(self, entered_input: str) -> EvaluatedFormatConstraint:
        """
        Assert that the entered input is the start/end of a german "Stromtag" (during central european daylight saving
        time).
        We ship this predefined method to evaluate the format constraints which are being introduced by expanding
        "time conditions" (UB1/UB3) in the :class:`expression_resolver.TimeConditionTransformer`.
        """
        return is_xtag_limit(entered_input, "Strom")

    # Attentive readers will notice, that 933 does exactly the same thing as 932; Also 935 does the same thing as 934.
    # This is because, from a data perspective, it's totally irrelevant if we're communicating a datetime with
    # +1h or +2h UTC offset as long as _any_ offset is given. Two distinctive format constraints are just ..., ok.
    # The authors of the AHBs probably had good intentions, when they introduced two different format constraints for
    # both German "winter" (CET/MEZ) and "summer time" (CEST/MESZ).
    # The road to hell is paved with good intentions.

    def evaluate_933(self, entered_input: str) -> EvaluatedFormatConstraint:
        """
        Assert that the entered input is the start/end of a german "Stromtag" (during central european standard time).
        We ship this predefined method to evaluate the format constraints which are being introduced by expanding
        "time conditions" (UB1/UB3) in the :class:`expression_resolver.TimeConditionTransformer`.
        """
        return is_xtag_limit(entered_input, "Strom")

    def evaluate_934(self, entered_input: str) -> EvaluatedFormatConstraint:
        """
        Assert that the entered input is the start/end of a german "Gastag" (during central european daylight saving
        time).
        We ship this predefined method to evaluate the format constraints which are being introduced by expanding
        "time conditions" (UB2/UB3) in the :class:`expression_resolver.TimeConditionTransformer`.
        """
        return is_xtag_limit(entered_input, "Gas")

    def evaluate_935(self, entered_input: str) -> EvaluatedFormatConstraint:
        """
        Assert that the entered input is the start/end of a german "Gastag" (during central european standard time).
        We ship this predefined method to evaluate the format constraints which are being introduced by expanding
        "time conditions" (UB2/UB3) in the :class:`expression_resolver.TimeConditionTransformer`.
        """
        return is_xtag_limit(entered_input, "Gas")

    async def evaluate_single_format_constraint(self, condition_key: str) -> EvaluatedFormatConstraint:
        """
        Evaluates the format constraint with the given key.
        :param condition_key: key of the condition, e.g. "950"
        :return: If the format constraint is fulfilled and an optional error_message.
        """
        text_to_be_evaluated = text_to_be_evaluated_by_format_constraint.get()
        evaluation_method = self.get_evaluation_method(condition_key)
        if evaluation_method is None:
            raise NotImplementedError(f"There is no content_evaluation method for format constraint '{condition_key}'")
        result: EvaluatedFormatConstraint
        if inspect.iscoroutinefunction(evaluation_method):
            result = await evaluation_method(text_to_be_evaluated)
        else:
            result = evaluation_method(text_to_be_evaluated)
        try:
            if result.format_constraint_fulfilled is False and result.error_message is None:
                result.error_message = f"Condition [{condition_key}] has to be fulfilled."
        except AttributeError as attribute_error:
            if isinstance(result, FormatConstraintEvaluationResult):
                # explicitly raise error with meaningful message, because this is really hard to distinguish for users
                raise ValueError(
                    "A FcEvaluator shall return EvaluatedFormatConstraints, _not_ FormatConstraintEvaluationResults"
                ) from attribute_error
            raise attribute_error
        self.logger.debug(
            "The format constraint %s with input '%s' evaluated to %s", condition_key, text_to_be_evaluated, str(result)
        )
        return result

    async def evaluate_format_constraints(self, condition_keys: List[str]) -> Dict[str, EvaluatedFormatConstraint]:
        """
        Evaluate the entered_input in regard to all the formats provided in condition_keys.
        """
        tasks: List[Coroutine] = [
            self.evaluate_single_format_constraint(condition_key) for condition_key in condition_keys
        ]
        results: List[EvaluatedFormatConstraint] = await asyncio.gather(*tasks)

        result: Dict[str, EvaluatedFormatConstraint] = dict(zip(condition_keys, results))
        return result


class DictBasedFcEvaluator(FcEvaluator):
    """
    A format constraint evaluator that is initialized with a prefilled dictionary on time of creation.
    Once initialized the outcome of the evaluation won't change anymore.
    """

    def __init__(self, results: Dict[str, EvaluatedFormatConstraint]):
        """
        Initialize with a dictionary that contains all the format constraint evaluation results.
        :param results:
        """
        super().__init__()
        self._results: Dict[str, EvaluatedFormatConstraint] = results

    # pylint: disable=unused-argument
    async def evaluate_single_format_constraint(self, condition_key: str) -> EvaluatedFormatConstraint:
        try:
            return self._results[condition_key]
        except KeyError as key_error:
            raise NotImplementedError(f"No result was provided for {condition_key}.") from key_error

    def get_evaluation_method(self, condition_key: str) -> Optional[Callable]:
        """
        Returns the method that evaluates the condition with key condition_key
        :param condition_key: unique key of the condition, e.g. "59"
        :return: The method that can be used for content_evaluation; None if no such method is implemented.
        """
        return self.get_evaluation_method(condition_key)


class ContentEvaluationResultBasedFcEvaluator(FcEvaluator):
    """
    A format constraint evaluator that expects the evaluatable data to contain a ContentEvalutionResult as edifact seed.
    Other than the DictBasedFcEvaluator the outcome is not dependent on the initialization but on the evaluatable data.
    """

    def __init__(self):
        super().__init__()
        self._schema = ContentEvaluationResultSchema()

    async def evaluate_single_format_constraint(self, condition_key: str) -> EvaluatedFormatConstraint:
        # the missing second argument to the private method call in the next line should be injected automatically
        return await self._evaluate_single_format_constraint(condition_key)  # pylint:disable=no-value-for-parameter

    @inject.params(evaluatable_data=EvaluatableDataProvider)  # injects what has been bound to the EvaluatableData type
    async def _evaluate_single_format_constraint(
        self, condition_key: str, evaluatable_data: EvaluatableData
    ) -> EvaluatedFormatConstraint:
        content_evaluation_result: ContentEvaluationResult = self._schema.load(evaluatable_data.body)
        try:
            self.logger.debug("Retrieving key %s' from Content Evaluation Result", condition_key)
            return content_evaluation_result.format_constraints[condition_key]
        except KeyError as key_error:
            raise NotImplementedError(f"No result was provided for {condition_key}.") from key_error

    def get_evaluation_method(self, condition_key: str) -> Optional[Callable]:
        async def evaluation_method():
            return await self.evaluate_single_format_constraint(condition_key)

        return evaluation_method



""" Tests the FC evaluator, that assumes a ContentEvaluationResult to be present in the evaluatable data"""
from typing import Optional

import inject
import pytest  # type:ignore[import]
from _pytest.fixtures import SubRequest  # type:ignore[import]

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult
from ahbicht.content_evaluation.token_logic_provider import TokenLogicProvider
from ahbicht.expressions.condition_nodes import EvaluatedFormatConstraint
from unittests.defaults import default_test_format, default_test_version


class TestCerBasedRcEvaluator:
    """Test for the evaluation using the ContentEvaluationResult Based FC Evaluator"""

    @pytest.mark.parametrize(
        "inject_cer_evaluators",
        [
            pytest.param(
                ContentEvaluationResult(
                    format_constraints={
                        "1": EvaluatedFormatConstraint(format_constraint_fulfilled=True),
                        "2": EvaluatedFormatConstraint(
                            format_constraint_fulfilled=False, error_message="something wrong"
                        ),
                    },
                    requirement_constraints={},
                    hints={},
                )
            )
        ],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "condition_key, expected_result",
        [
            pytest.param("1", EvaluatedFormatConstraint(format_constraint_fulfilled=True)),
            pytest.param(
                "2",
                EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="something wrong"),
            ),
            pytest.param(
                "789",
                None,
            ),
        ],
    )
    async def test_evaluation(
        self, condition_key: str, expected_result: Optional[EvaluatedFormatConstraint], inject_cer_evaluators
    ):
        token_logic_provider: TokenLogicProvider = inject.instance(TokenLogicProvider)  # type:ignore[assignment]
        fc_evalutor = token_logic_provider.get_fc_evaluator(default_test_format, default_test_version)
        if expected_result is not None:
            actual = await fc_evalutor.evaluate_single_format_constraint(condition_key)
            assert actual == expected_result
        else:
            with pytest.raises(NotImplementedError):
                await fc_evalutor.evaluate_single_format_constraint(condition_key)



"""
Dataclasses that are relevant in the context of the content_evaluation.
"""
from dataclasses import dataclass, replace
from typing import Generic, Optional, TypeVar

from maus.edifact import EdifactFormat, EdifactFormatVersion

_BodyT = TypeVar("_BodyT")
"""
the type of the data on which the evaluations are performed
"""


@dataclass
class EvaluatableData(Generic[_BodyT]):
    """
    Data that can be processed/evaluated by an evaluator. They must not change during a content_evaluation run.
    The evaluatable data act as a flexible container to pass information that might be necessary for the
    content_evaluation. As of now (because our content_evaluation capabilities are quite limited) the only information
    provided is the meta seed of the message itself. But in the future the data provided might grow.
    """

    body: _BodyT  #: the body of the message that is being validated in the respective format (e.g. an edifact_seed)
    edifact_format: EdifactFormat  #: the format of the evaluatable message (e.g. UTILMD)
    edifact_format_version: EdifactFormatVersion  #: the format version of the evaluable data (e.g. FV2210)
    # ideas for what else could go here:
    # - pruefidentifikator to tweak the content_evaluation depending on the situation?


# pylint:disable=too-few-public-methods
class EvaluatableDataProvider(Generic[_BodyT]):
    """
    This is just a dummy class that is used for dependency injection.
    Use it to call binder.bind_to_provider(EvaluatableDataProvider, func_that_returns_evaluatable_data_goes_here)
    during dependency injection.
    See https://github.com/ivankorobkov/python-inject#why-no-scopes
    """


@dataclass
class EvaluationContext:
    """
    A content_evaluation context describes the setting in which a condition shall be evaluated. The content_evaluation
    context might have different values for the same condition in one content_evaluation run. E.g. if the purpose of the
    condition is to make sure that every Zähler with zähler type "EHZ" has some properties the context of the
    content_evaluation is one zähler entry although there might be multiple zählers present in the message.
    """

    scope: Optional[
        str
    ]  # jsonpath that refers to the scope of the content_eval. If None, then "$" = entire message is used as scope.


def copy_evaluation_context(context: EvaluationContext) -> EvaluationContext:
    """
    Returns a deep copy of the provided context.
    This allows you to create a copy of a context instead of modifying the original context (as EvaluationContexts are
    "pass by reference")
    :param context:
    :return: a deep copy of the context
    """
    return replace(context)



""" Tests the RC evaluator, that assumes a ContentEvaluationResult to be present in the evaluatable data"""
from unittest import mock

import pytest  # type:ignore[import]

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult
from ahbicht.content_evaluation.rc_evaluators import ContentEvaluationResultBasedRcEvaluator, RcEvaluator
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue
from unittests.conftest import store_content_evaluation_result_in_evaluatable_data


class TestCerBasedRcEvaluator:
    """Test for the evaluation using the ContentEvaluationResult Based RC Evaluator"""

    async def test_evaluation(self, mocker):
        hardcoded_results = ContentEvaluationResult(
            format_constraints={},
            hints={},
            requirement_constraints={
                "1": ConditionFulfilledValue.NEUTRAL,
                "2": ConditionFulfilledValue.UNFULFILLED,
                "3": ConditionFulfilledValue.FULFILLED,
                "4": ConditionFulfilledValue.UNKNOWN,
            },
        )
        evaluator: RcEvaluator = ContentEvaluationResultBasedRcEvaluator()
        dummy_eval_data = store_content_evaluation_result_in_evaluatable_data(
            content_evaluation_result=hardcoded_results
        )
        assert (
            await evaluator.evaluate_single_condition("1", evaluatable_data=dummy_eval_data)
            == ConditionFulfilledValue.NEUTRAL
        )
        assert (
            await evaluator.evaluate_single_condition("2", evaluatable_data=dummy_eval_data)
            == ConditionFulfilledValue.UNFULFILLED
        )
        assert (
            await evaluator.evaluate_single_condition("3", evaluatable_data=dummy_eval_data)
            == ConditionFulfilledValue.FULFILLED
        )
        assert (
            await evaluator.evaluate_single_condition("4", evaluatable_data=dummy_eval_data)
            == ConditionFulfilledValue.UNKNOWN
        )
        with pytest.raises(NotImplementedError):
            await evaluator.evaluate_single_condition("5", evaluatable_data=dummy_eval_data)

        single_condition_spy = mocker.spy(evaluator, "evaluate_single_condition")

        assert (
            await evaluator.evaluate_conditions(["1", "2", "3", "4"], evaluatable_data=dummy_eval_data)
            == hardcoded_results.requirement_constraints
        )
        single_condition_spy.assert_has_awaits(
            [
                mock.call("1", dummy_eval_data, None),
                mock.call("2", dummy_eval_data, None),
                mock.call("3", dummy_eval_data, None),
                mock.call("4", dummy_eval_data, None),
            ]
        )
